@@ -37,6 +37,10 @@
 
 #include "exmap.h"
 
+#ifdef INTERACTIVE
+#include <readline/readline.h>
+#endif
+
 #define EXIT_KERNEL_ERROR  -1
 #define EXIT_LOGFILE_ERROR -2
 #define EXIT_SYSDATA_ERROR -3
@@ -48,12 +52,12 @@
 
 #if 1
 #define SCALE_V(x) \
-    x > 1073741824 ? x >> 20 : \
-    x > 1048576 ? x >> 10 : x
+    x > 104857600 ? x >> 20 : \
+    x > 102400 ? x >> 10 : x
 
 #define SCALE_U(x) \
-    x > 1073741824 ? "MB" : \
-    x > 1048576 ? "KB" : " B"
+    x > 104857600 ? "MB" : \
+    x > 102400 ? "KB" : " B"
 #else
 #define SCALE_V(x) x
 #define SCALE_U(x) "B"
@@ -66,20 +70,36 @@ extern int columns;
 
 static data_t data;
 
-static void
+static gboolean
 myprintf (data_t *d, const char *fmt, ...)
 {
   va_list args;
 
   if (! (d && d->of && fmt))
-    return;
+    return FALSE;
 
 #ifdef INTERACTIVE
   if (d->args.interactive && d->global_lines >= rows - 1)
     {
+      char * line;
+      gboolean done = FALSE;
+      
       d->global_lines = 0;
-      fprintf (d->of, "  --- Press Enter for more ---");
-      fgetc (stdin);
+
+      /* read single character from kbd; reset back to reading
+       * entire line afterwards
+       */
+      rl_num_chars_to_read = 1;
+      line = readline ("  --- Press Return for more or q to quit ---");
+      rl_num_chars_to_read = 0;
+      
+      if (*line == 'q' || *line == 'Q')
+        done = TRUE;
+
+      free (line);
+
+      if (done)
+        return FALSE;
     }
 #endif
   
@@ -88,9 +108,11 @@ myprintf (data_t *d, const char *fmt, ...)
   va_end (args);
 
   d->global_lines++;
+
+  return TRUE;
 }
 
-guint
+gint
 print_paged (data_t * d, char * text)
 {
   if (!d || !text)
@@ -107,7 +129,9 @@ print_paged (data_t * d, char * text)
       ++count;
       *nl = 0;
 
-      myprintf (d, "%s\n", t);
+      if (!myprintf (d, "%s\n", t))
+        return -1;
+      
       t = nl + 1;
       
       nl = strchr (t, '\n');
@@ -116,7 +140,9 @@ print_paged (data_t * d, char * text)
 
   if (*t)
     {
-      myprintf (d, "%s\n", t);
+      if (!myprintf (d, "%s\n", t))
+        return -1;
+      
       ++count;
     }
 
@@ -164,10 +190,17 @@ memory_destroy (memory_t * m)
     return;
   
   if (m->map)
-    g_hash_table_destroy (m->map);
-
+    {
+      g_hash_table_destroy (m->map);
+      m->map = NULL;
+    }
+  
   if (m->files)
-    g_hash_table_destroy (m->files);
+    {
+      g_hash_table_destroy (m->files);
+      m->files = NULL;
+    }
+  
   
   g_free (m->cmd);
 }
@@ -251,14 +284,13 @@ file_free (file_t * f)
  * a memory struct for this particular pid.
  */
 static void
-process_vma (gpointer key, gpointer value, gpointer data)
+process_vma (gpointer key, gpointer value, gpointer user_data)
 {
   vma_t    * v  = (vma_t*) value;
-  memory_t * m  = (memory_t*)data;
+  memory_t * m  = (memory_t*)user_data;
   GList    * l  = v->pages;
   file_t * fm = NULL;
   
-
   if (v->file)
     {
       if (!(fm = g_hash_table_lookup (m->files, v->file)))
@@ -267,7 +299,11 @@ process_vma (gpointer key, gpointer value, gpointer data)
 
           /* store the file name in the cmd field, it comes handy later */
           fm->m.cmd = g_strdup (v->file);
-          fm->elf = elf_load (v->file, m->load_symbols);
+
+          /* Evil global variable ... */
+          if (data.srv_sckt < 0 || data.args.fs_root)
+            fm->elf = elf_load (v->file, m->load_symbols);
+          
           fm->m.pg_count = m->pg_count;
           
           g_hash_table_insert (m->files, g_strdup (v->file), fm);
@@ -474,7 +510,7 @@ process_files (gpointer key, gpointer value, gpointer data)
     }
 }
 
-static guint
+static gint
 print_symbol_info (symbol_t * s, data_t * d)
 {
   if (!s || !s->m.vm)
@@ -482,24 +518,26 @@ print_symbol_info (symbol_t * s, data_t * d)
 
   guint lines = 0;
   
-  if (d->args.format == FORMAT_HUMAN)
+  if (d->args.format == FORMAT_HUMAN || d->args.interactive)
     {
-      myprintf (d, "      %-26s: VM%8u %s, M%8u %s, S%8u %s\n",
-               s->name,
-               SCALE_V(s->m.vm),     SCALE_U(s->m.vm),
-               SCALE_V(s->m.mapped), SCALE_U(s->m.mapped),
-               SCALE_V(s->m.sole),   SCALE_U(s->m.sole));
+      if (!myprintf (d, "      %-26s: VM%8u %s, M%8u %s, S%8u %s\n",
+                     s->name,
+                     SCALE_V(s->m.vm),     SCALE_U(s->m.vm),
+                     SCALE_V(s->m.mapped), SCALE_U(s->m.mapped),
+                     SCALE_V(s->m.sole),   SCALE_U(s->m.sole)))
+        return -1;
       
       lines++;
     }
   else if (d->args.format == FORMAT_CSV || d->args.format == FORMAT_CSV_SUMS)
     {
-      myprintf (d, ",,,,,,,,,,,,\"%s\",%u,%u,%u\n",
-               s->name, 
-               s->m.vm,
-               s->m.mapped,
-               s->m.sole);
-
+      if (!myprintf (d, ",,,,,,,,,,,,\"%s\",%u,%u,%u\n",
+                     s->name, 
+                     s->m.vm,
+                     s->m.mapped,
+                     s->m.sole))
+        return -1;
+      
       lines++;
     }
 
@@ -507,7 +545,7 @@ print_symbol_info (symbol_t * s, data_t * d)
 }
 
 /* TODO refactor these print fuctions to share code */
-static guint
+static gint
 print_section_info (section_t * s, data_t * d)
 {
   if (!s)
@@ -517,39 +555,45 @@ print_section_info (section_t * s, data_t * d)
 
   if (s->m.vm)
     {
-      if (d->args.format == FORMAT_HUMAN)
+      if (d->args.format == FORMAT_HUMAN || d->args.interactive)
         {
-          myprintf (d,"   %-29s: VM%8u %s, M%8u %s, S%8u %s\n",
-                   s->name,
-                   SCALE_V(s->m.vm),     SCALE_U(s->m.vm),
-                   SCALE_V(s->m.mapped), SCALE_U(s->m.mapped),
-                   SCALE_V(s->m.sole),   SCALE_U(s->m.sole));
+          if (!myprintf (d,"   %-29s: VM%8u %s, M%8u %s, S%8u %s\n",
+                         s->name,
+                         SCALE_V(s->m.vm),     SCALE_U(s->m.vm),
+                         SCALE_V(s->m.mapped), SCALE_U(s->m.mapped),
+                         SCALE_V(s->m.sole),   SCALE_U(s->m.sole)))
+            return -1;
       
           lines++;
         }
       else if (d->args.format == FORMAT_CSV ||
                d->args.format == FORMAT_CSV_SUMS)
         {
-          myprintf (d,",,,,,,,,,,,,\"%s\",%u,%u,%u\n",
-                   s->name, 
-                   s->m.vm,
-                   s->m.mapped,
-                   s->m.sole);
+          if (!myprintf (d,",,,,,,,,,,,,\"%s\",%u,%u,%u\n",
+                         s->name, 
+                         s->m.vm,
+                         s->m.mapped,
+                         s->m.sole))
+            return -1;
 
           lines++;
         }
     }
   else if (s->syms)
     {
-      if (d->args.format == FORMAT_HUMAN)
+      if (d->args.format == FORMAT_HUMAN || d->args.interactive)
         {
-          myprintf (d,"   %-29s:\n", s->name);
+          if (!myprintf (d,"   %-29s:\n", s->name))
+            return -1;
+          
           lines++;
         }
       else if (d->args.format == FORMAT_CSV ||
                d->args.format == FORMAT_CSV_SUMS)
         {
-          myprintf (d,",,,,,,,,,,,,\"%s\",0,0,0\n", s->name);
+          if (!myprintf (d,",,,,,,,,,,,,\"%s\",0,0,0\n", s->name))
+            return -1;
+          
           lines++;
         }
     }
@@ -558,14 +602,19 @@ print_section_info (section_t * s, data_t * d)
   while (l)
     {
       symbol_t * sbl = l->data;
-      lines += print_symbol_info (sbl, d);
+      int i = print_symbol_info (sbl, d);
+
+      if (i < 0)
+        return -1;
+      
+      lines += i;
       l = l->next;
     }
   
   return lines;
 }
 
-static guint
+static gint
 print_file_info (file_t * f, data_t * d)
 {
   if (!f)
@@ -574,20 +623,24 @@ print_file_info (file_t * f, data_t * d)
   guint lines = 0;
   gchar * cmd = g_path_get_basename (f->m.cmd);
   
-  if (d->args.format == FORMAT_HUMAN)
+  if (d->args.format == FORMAT_HUMAN || d->args.interactive)
     {
       if (d->args.detail & DETAIL_ELF_INFO)
         {
-          myprintf (d,"  ----------------------------------"
-                   "---------------------------------------\n");
+          if (!myprintf (d,"  ----------------------------------"
+                         "---------------------------------------\n"))
+            return -1;
+          
           lines++;
         }
 
-      myprintf (d,"  %-30s: VM%8u %s, M%8u %s, S%8u %s ",
-               cmd,
-               SCALE_V(f->m.vm),     SCALE_U(f->m.vm),
-               SCALE_V(f->m.mapped), SCALE_U(f->m.mapped),
-               SCALE_V(f->m.sole),   SCALE_U(f->m.sole));
+      if (!myprintf (d,"  %-30s: VM%8u %s, M%8u %s, S%8u %s ",
+                     cmd,
+                     SCALE_V(f->m.vm),     SCALE_U(f->m.vm),
+                     SCALE_V(f->m.mapped), SCALE_U(f->m.mapped),
+                     SCALE_V(f->m.sole),   SCALE_U(f->m.sole)))
+        return -1;
+      
       d->global_lines--;
       
 #if 0
@@ -596,23 +649,27 @@ print_file_info (file_t * f, data_t * d)
       if (d->args.detail & DETAIL_ELF_INFO)
         {
           lines += 2;
-          myprintf (d,"\n  ----------------------------------"
-                   "---------------------------------------\n");
+          if (!myprintf (d,"\n  ----------------------------------"
+                         "---------------------------------------\n"))
+            return -1;
+          
           d->global_lines++;
         }
       else
         {
           lines++;
-          myprintf (d,"\n");
+          if (!myprintf (d,"\n"))
+            return -1;
         }
     }
   else if (d->args.format == FORMAT_CSV || d->args.format == FORMAT_CSV_SUMS)
     {
-      myprintf (d,",,,,,,,,\"%s\",%u,%u,%u\n",
-               f->m.cmd, 
-               f->m.vm,
-               f->m.mapped,
-               f->m.sole);
+      if (!myprintf (d,",,,,,,,,\"%s\",%u,%u,%u\n",
+                     f->m.cmd, 
+                     f->m.vm,
+                     f->m.mapped,
+                     f->m.sole))
+        return -1;
 
       lines++;
     }
@@ -625,7 +682,12 @@ print_file_info (file_t * f, data_t * d)
       while (l)
         {
           section_t * s= (section_t*)l->data;
-          lines += print_section_info (s, d);
+          int i = print_section_info (s, d);
+
+          if (i < 0)
+            return -1;
+          
+          lines += i;
           l = l->next;
         }
     }
@@ -859,11 +921,17 @@ load_map (data_t * d, pid_t pid)
 
               if (n)
                 *n = 0;
-#if 0
-              v->file = g_path_get_basename (b);
-#else
-              v->file = g_strdup (b);
-#endif
+
+              /* If we are not connected to remote sever, or no FS root
+               * has been specified, we simply copy the file name.
+               *
+               * If we both have FS root and are connected, we prefix the
+               * file name with the root
+               */
+              if (!d->args.fs_root || d->srv_sckt < 0)
+                v->file = g_strdup (b);
+              else
+                v->file = g_build_filename (d->args.fs_root, b, NULL);
 	    }
 	}
       else
@@ -1186,7 +1254,7 @@ sort_pid_list (gconstpointer a, gconstpointer b, gpointer data)
 /* Prints contents of a memory struct to stdout.
  * file_info indicates if per-file information is to be printed as well
  */
-static guint
+static gint
 print_results (memory_t * mem_data, data_t * d)
 {
   guint lines = 0;
@@ -1197,32 +1265,55 @@ print_results (memory_t * mem_data, data_t * d)
   if (threshold && mem_data->vm < threshold)
     return lines;
 
-  if (d->args.format == FORMAT_HUMAN)
+  if (d->args.format == FORMAT_HUMAN || d->args.interactive)
     {
-      myprintf (d,"Process %d", mem_data->pid);
+      if (!myprintf (d,"Process %d", mem_data->pid))
+        return -1;
+      
       d->global_lines--;
       
       if (mem_data->cmd)
-        myprintf (d," [%s]\n", mem_data->cmd);
+        {
+          if (!myprintf (d," [%s]\n", mem_data->cmd))
+            return -1;
+        }
       else
-        myprintf (d,"\n");
-  
-      myprintf (d," Virtual memory  : %9u %s\n",
-               SCALE_V(mem_data->vm), SCALE_U(mem_data->vm));
-      myprintf (d," Effective VM    : %9u %s\n",
-               SCALE_V(mem_data->evm), SCALE_U(mem_data->evm));
-      myprintf (d," Heap            : %9u %s\n",
-               SCALE_V(mem_data->heap), SCALE_U(mem_data->heap));
-      myprintf (d," Stack           : %9u %s\n",
-               SCALE_V(mem_data->stack), SCALE_U(mem_data->stack));
-      myprintf (d," vdso            : %9u %s\n",
-               SCALE_V(mem_data->vdso), SCALE_U(mem_data->vdso));
-      myprintf (d," Mapped          : %9u %s\n",
-               SCALE_V(mem_data->mapped), SCALE_U(mem_data->mapped));
-      myprintf (d," Effective mapped: %9u %s\n",
-               SCALE_V(mem_data->emapped), SCALE_U(mem_data->emapped));
-      myprintf (d," Sole use        : %9u %s\n",
-               SCALE_V(mem_data->sole), SCALE_U(mem_data->sole));
+        {
+          if (!myprintf (d,"\n"))
+            return -1;
+        }
+      
+      if (!myprintf (d," Virtual memory  : %9u %s\n",
+                     SCALE_V(mem_data->vm), SCALE_U(mem_data->vm)))
+        return -1;
+      
+      if (!myprintf (d," Effective VM    : %9u %s\n",
+                     SCALE_V(mem_data->evm), SCALE_U(mem_data->evm)))
+        return -1;
+      
+      if (!myprintf (d," Heap            : %9u %s\n",
+                     SCALE_V(mem_data->heap), SCALE_U(mem_data->heap)))
+        return -1;
+      
+      if (!myprintf (d," Stack           : %9u %s\n",
+                     SCALE_V(mem_data->stack), SCALE_U(mem_data->stack)))
+        return -1;
+      
+      if (!myprintf (d," vdso            : %9u %s\n",
+                     SCALE_V(mem_data->vdso), SCALE_U(mem_data->vdso)))
+        return -1;
+      
+      if (!myprintf (d," Mapped          : %9u %s\n",
+                     SCALE_V(mem_data->mapped), SCALE_U(mem_data->mapped)))
+          return -1;
+          
+      if (!myprintf (d," Effective mapped: %9u %s\n",
+                     SCALE_V(mem_data->emapped), SCALE_U(mem_data->emapped)))
+        return -1;
+      
+      if (!myprintf (d," Sole use        : %9u %s\n",
+                     SCALE_V(mem_data->sole), SCALE_U(mem_data->sole)))
+        return -1;
 
       lines += 9;
       
@@ -1230,24 +1321,26 @@ print_results (memory_t * mem_data, data_t * d)
         {
           lines += 2;
           d->global_lines++;
-          myprintf (d,"\n Per file memory use\n");
+          if (!myprintf (d,"\n Per file memory use\n"))
+            return -1;
         }
     }
   else if (d->args.format == FORMAT_CSV || d->args.format == FORMAT_CSV_SUMS)
     {
       lines++;
       
-      myprintf (d,"%u,\"%s\",%u,%u,%u,%u,%u,%u,%u,%u\n",
-               mem_data->pid,
-               mem_data->cmd ? mem_data->cmd : "???",
-               mem_data->vm,
-               mem_data->evm,
-               mem_data->heap,
-               mem_data->stack,
-               mem_data->vdso,
-               mem_data->mapped,
-               mem_data->emapped,
-               mem_data->sole);
+      if (!myprintf (d,"%u,\"%s\",%u,%u,%u,%u,%u,%u,%u,%u\n",
+                     mem_data->pid,
+                     mem_data->cmd ? mem_data->cmd : "???",
+                     mem_data->vm,
+                     mem_data->evm,
+                     mem_data->heap,
+                     mem_data->stack,
+                     mem_data->vdso,
+                     mem_data->mapped,
+                     mem_data->emapped,
+                     mem_data->sole))
+        return -1;
     }
 
   if (d->args.detail & DETAIL_FILE_INFO)
@@ -1257,28 +1350,41 @@ print_results (memory_t * mem_data, data_t * d)
 
       if (l)
         {
-          l = g_list_sort_with_data (l, sort_pid_list,
-                                     GINT_TO_POINTER (d->args.sort));
+          GList * list = g_list_sort_with_data (l, sort_pid_list,
+                                              GINT_TO_POINTER (d->args.sort));
+          l = list;
+          
           while (l)
             {
-              lines += print_file_info ((file_t*)l->data, d);
+              int i = print_file_info ((file_t*)l->data, d);
+
+              if (i < 0)
+                {
+                  g_list_free (list);
+                  return -1;
+                }
+              
+              lines += i;
               l = l->next;
             }
 
-          g_list_free (l);
+          g_list_free (list);
         }
     }
 
-  if (d->args.format == FORMAT_HUMAN)
+  if (d->args.format == FORMAT_HUMAN || d->args.interactive)
     {
       if (!d->args.interactive)
         {
-          myprintf (d,"\n========================================"
-                       "=======================================\n\n");
+          if (!myprintf (d,"\n========================================"
+                         "=======================================\n\n"))
+            return -1;
+          
           d->global_lines += 2;
         }
       else
-        myprintf (d, "\n");
+        if (!myprintf (d, "\n"))
+          return -1;
     }
 
   return lines;
@@ -1315,13 +1421,14 @@ usage (const char * option)
   fprintf (f, "  --no-elf-info         : No breakdown per elf sections in each file.\n");
   fprintf (f, "  --all-pids            : Information for all accessible pids; if not used,\n"
               "                          at least one process must be specified.\n");
-  fprintf (f, "  --vm-threshold=bytes  : Information only for processes above VM threshold\n");
+  fprintf (f, "  --vm-threshold=size   : Information only for processes above VM threshold\n");
   fprintf (f, "  --sort-by=what        : vm|sole|heap|stack|mapped|none (default: none)\n");
   fprintf (f, "  --sort-type=which     : desc|asc (default: desc)\n");
   fprintf (f, "  --format=which        : human|csv|csv+sums (default: human)\n");
   fprintf (f, "  --log=file            : Print results to file, not stdout\n");
   fprintf (f, "  --interactive         : Run in interactive mode\n");
   fprintf (f, "  --remote              : Run in interactive mode for remote connection\n");
+  fprintf (f, "  --root                : Root of the file system to use (remote session only)\n");
   fprintf (f, "  --version             : Program version\n");
   fprintf (f, "  --module-version      : Version of required kernel module\n");
   fprintf (f, "  --help\n\n");
@@ -1334,6 +1441,37 @@ usage (const char * option)
   fprintf (f, "processes specified explicitely symbol information is printed where available.\n\n");
 }
 
+gint
+get_unit (const char * p)
+{
+  gint multiplier = 1;
+  
+  while (*p && (isdigit (*p)|| isspace(*p)))
+    ++p;
+
+  switch (*p)
+    {
+    case 'g':
+    case 'G':
+      multiplier *= 1024;
+      /* fall through */
+    case 'm':
+    case 'M':
+      multiplier *= 1024;
+      /* fall through */
+    case 'k':
+    case 'K':
+      multiplier *= 1024;
+      break;
+
+    default:
+      ;
+    }
+
+  return multiplier;
+}
+
+
 static gint
 parse_args (int argc, char ** argv, args_t * args)
 {
@@ -1342,7 +1480,7 @@ parse_args (int argc, char ** argv, args_t * args)
   if (argc < 2)
     {
       usage (NULL);
-      return 0;
+      return EXIT_INVALID_ARGS;
     }
 
   for (i = 1; i < argc; ++i)
@@ -1353,6 +1491,7 @@ parse_args (int argc, char ** argv, args_t * args)
       if (!strcmp (argv[i], "--no-file-info"))
         {
           DETAIL_UNSET (args->detail, DETAIL_FILE_INFO);
+          DETAIL_UNSET (args->detail, DETAIL_ELF_INFO);
           continue;
         }
 
@@ -1401,7 +1540,8 @@ parse_args (int argc, char ** argv, args_t * args)
 
       if (!strncmp (argv[i], "--vm-threshold",14))
         {
-          args->threshold = atoi (argv[i] + 15);
+          gint unit = get_unit (argv[i] + 15);
+          args->threshold = unit * atoi (argv[i] + 15);
           continue;
         }
 
@@ -1446,6 +1586,18 @@ parse_args (int argc, char ** argv, args_t * args)
           return EXIT_INVALID_ARGS;
 #endif
         }
+
+      if (!strncmp (argv[i], "--root", 6))
+        {
+#ifdef INTERACTIVE
+          args->fs_root = g_strdup (argv[i] + 7);
+          continue;
+#else
+          g_print ("--root not supported "
+                   "(compiled with --without-readline)\n");
+          return EXIT_INVALID_ARGS;
+#endif
+        }
       
       if (!strcmp (argv[i], "--help"))
         {
@@ -1643,34 +1795,49 @@ main_print (data_t * d)
   /* output header */
   if (d->args.format == FORMAT_CSV || d->args.format == FORMAT_CSV_SUMS)
     {
-      myprintf (d,
-               "pid,command,VM,effective VM,heap,stack,vdso,mapped,"
-               "emapped,sole");
+      if (!myprintf (d,
+                     "pid,command,VM,effective VM,heap,stack,vdso,mapped,"
+                     "emapped,sole"))
+        return;
+      
       d->global_lines--;
       
       if (d->args.detail & DETAIL_FILE_INFO)
         {
-          myprintf (d,",file,file-VM,file-mapped,file-solefile");
+          if (!myprintf (d,",file,file-VM,file-mapped,file-solefile"))
+            return;
 
           if (d->args.detail & DETAIL_ELF_INFO)
-            myprintf (d,
-                     ",section,section-VM,section-mapped,section-sole\n");
+            {
+              if (!myprintf (d,
+                          ",section,section-VM,section-mapped,section-sole\n"))
+                return;
+            }
           else
-            myprintf (d,"\n");
-
+            {
+              if (!myprintf (d,"\n"))
+                return;
+            }
+          
           d->global_lines--;
         }
       else
-        myprintf (d,"\n");
+        if (!myprintf (d,"\n"))
+          return;
     }
 
   /* output info for individula pids */
-  if (d->args.all_pids)
+  if (d->args.all_pids || !d->args.pids_arg || !*d->args.pids_arg)
     {
       l = d->mem_list;
       while (l)
         {
-          lines += print_results ((memory_t*)l->data, d);
+          int i = print_results ((memory_t*)l->data, d);
+
+          if (i < 0)
+            return;
+          
+          lines += i;
           l = l->next;
         }
     }
@@ -1685,7 +1852,15 @@ main_print (data_t * d)
                                       GINT_TO_POINTER (*pid), find_pid);
 
               if (l)
-                lines += print_results ((memory_t*)l->data, d);
+                {
+                  int i = print_results ((memory_t*)l->data, d);
+
+                  if (i < 0)
+                    return;
+                  
+                  lines += i;
+                }
+              
               pid++;
             }
         }
@@ -1758,10 +1933,10 @@ main_reset (data_t * d)
     }
 }
 
-#ifdef INTERACTIVE
 static void
 signal_handler (int sig)
 {
+#ifdef INTERACTIVE
   if (sig == SIGPIPE)
     {
       g_print ("Connection to remote server lost.\n");
@@ -1773,8 +1948,21 @@ signal_handler (int sig)
           data.srv_sckt = -1;
         }
     }
-}
 #endif
+  
+  /* If we have an open socket to a remote server, close it */
+  if (sig == SIGTERM || sig == SIGSEGV || sig == SIGABRT)
+    {
+#ifdef INTERACTIVE
+      if (data.srv_sckt >= 0)
+        close (data.srv_sckt);
+
+      interactive_save_history (&data);
+#endif
+      g_message ("caught signal %d, exiting", sig);
+      exit (sig);
+    }
+}
 
 int
 main (int argc, char ** argv)
@@ -1800,13 +1988,16 @@ main (int argc, char ** argv)
       goto finish;
     }
 
-#ifdef INTERACTIVE
-  /* handle SIGPIPE */
+  /* handle signals */
   struct sigaction sa;
   sigfillset(&sa.sa_mask);
   sa.sa_handler = signal_handler;
+#ifdef INTERACTIVE
   sigaction(SIGPIPE, &sa, NULL);
 #endif
+  sigaction(SIGTERM, &sa, NULL);
+  sigaction(SIGSEGV, &sa, NULL);
+  sigaction(SIGABRT, &sa, NULL);
   
 #ifdef INTERACTIVE
   if (data.args.interactive)
