@@ -148,6 +148,7 @@ typedef struct
   guint         threshold;
   format_t      format;
   const char  * log;
+  char        * fs_root;
 } args_t;
 
 typedef struct
@@ -162,7 +163,8 @@ typedef struct
 }data_t;
 
 
-guint    print_paged (data_t * d, char * text);
+gint     print_paged (data_t * d, char * text);
+gint     get_unit (const char * p);
 
 void     main_load (data_t * d);
 void     main_reset (data_t * d);
@@ -172,5 +174,6 @@ void     main_print (data_t * d);
 void     interactive_init (data_t * d);
 void     interactive_loop (data_t * d);
 gboolean interactive_parse_line (data_t *d, char * line);
+void     interactive_save_history (data_t *d);
 
 #endif /* ifndef EXMAP_H */
