@@ -418,9 +418,8 @@ interactive_connect (data_t * d, const char * server)
 
   /* Handshake */
   char buf[100];
-  g_snprintf (buf, sizeof (buf),
-              "EXMP_HANDSHAKE %s-%s\n",
-              VERSION, EXMAP_VERSION);
+  g_snprintf (buf, sizeof (buf), "EXMP_HANDSHAKE %s\n",
+              EXMAP_PROTOCOL_VERSION);
   
   write (d->srv_sckt, buf, strlen (buf));
 
@@ -461,6 +460,25 @@ interactive_connect (data_t * d, const char * server)
   return EMSG_SUCCESS;
 }
 
+static guint
+interactive_root (data_t * d, const char * r)
+{
+  if (d->args.fs_root)
+    g_free (d->args.fs_root);
+
+  if (r && *r)
+    d->args.fs_root = g_strdup (r);
+  else
+    d->args.fs_root = NULL;
+
+  main_reset (d);
+
+  if (d->srv_sckt >= 0)
+    main_load (d);
+
+  return EMSG_SUCCESS;
+}
+
 static guint
 interactive_help ()
 {
@@ -468,26 +486,46 @@ interactive_help ()
       "Available commands:\n"
       "  add 'all'|pid|command       : select process for examination\n"
       "  clear                       : deselect all selected processes\n"
-      "  connect 'host:port'         : connect to remote exmapserver\n"
+      "  connect host:port           : connect to remote exmap server\n"
       "  debug                       : some debug info\n"
       "  detail 'none'|'file'|'elf'  : set level of detail to show\n"
       "  disconnect                  : disconnect from remote server\n"
+      "  format 'human'|'csv'        : format for saving data\n"
       "  ls                          : list selected processes\n"
       "  print ['all'|pid|command]   : print data for selected/specified processes\n"
       "  ps [options]                : run system ps command\n"
       "  reload                      : recreate system snapshot\n"
       "  rm pid|'all'                : deselect process with pid, or clear 'all' flag\n"
+      "  root fsroot                 : change file system root for remote session\n"
       "  save file                   : save data for selected process to file\n"
       "  shell commands              : run shell commands\n"
       "  sort asc|desc|vm|sole|heap|stack|mapped|emapped\n"
       "                              : sort results by\n"
-      "  threshold                   : set minimum VM for processes printed\n"
+      "  threshold size [k|M|G]      : set minimum VM for processes printed\n"
       "  ?, help                     : help\n"
       "  quit                        : quit\n");
 
   return EMSG_SUCCESS;
 }
 
+static guint
+interactive_format (data_t * d, const char *p)
+{
+  if (!p)
+    return EMSG_INVALID_PARAMETER;
+
+  if (!strncmp (p, "human", 5))
+    d->args.format = FORMAT_HUMAN;
+  else if (!strncmp (p, "csv", 3))
+    d->args.format = FORMAT_CSV;
+  else if (!strncmp (p, "csv+sums", 8))
+    d->args.format = FORMAT_CSV_SUMS;
+  else
+    return EMSG_INVALID_PARAMETER;
+
+  return EMSG_SUCCESS;
+}
+
 static guint
 interactive_debug (data_t * d)
 {
@@ -495,6 +533,19 @@ interactive_debug (data_t * d)
   return EMSG_SUCCESS;
 }
 
+static guint
+interactive_threshold (data_t * d, const char *p)
+{
+  int unit = get_unit (p);
+  
+  if (*p)
+    d->args.threshold = unit * atoi (p);
+  else
+    return EMSG_INVALID_PARAMETER;
+
+  return EMSG_SUCCESS;
+}
+
 static guint
 interactive_spawn (data_t *d, const char * cmd, const char * params)
 {
@@ -585,6 +636,7 @@ interactive_parse_line (data_t *d, char * line)
   gboolean   history = TRUE;
   guint      retval = EMSG_SUCCESS;
   gchar    * cmd;
+  gchar    * param;
   size_t     cmd_len;
 
   if (!line)
@@ -607,13 +659,21 @@ interactive_parse_line (data_t *d, char * line)
       p++;
     }
 
+  param = p;
+
+  while (*p && !isspace (*p))
+    p++;
+
+  if (*p)
+    *p = 0;
+  
   cmd_len = strlen (cmd);
 
   switch (c)
     {
     case 'a':
       if (!strncmp (cmd, "add", cmd_len))
-        retval = interactive_add (d, p);
+        retval = interactive_add (d, param);
       else
         retval = EMSG_UNKNOWN_COMMAND;
       break;
@@ -628,7 +688,7 @@ interactive_parse_line (data_t *d, char * line)
         }
       else if (!strncmp (cmd, "connect", cmd_len))
         {
-          retval = interactive_connect (d, p);
+          retval = interactive_connect (d, param);
         }
       else
         retval = EMSG_UNKNOWN_COMMAND;
@@ -637,15 +697,26 @@ interactive_parse_line (data_t *d, char * line)
     case 'd':
       if (!strncmp (cmd, "detail", cmd_len))
         {
-          if (!strcmp (p, "none"))
+          if (!strcmp (param, "none"))
             d->args.detail = 0;
-          else if (!strcmp (p, "file"))
-            DETAIL_SET (d->args.detail, DETAIL_FILE_INFO);
-          else if (!strcmp (p, "elf"))
-            DETAIL_SET (d->args.detail, DETAIL_ELF_INFO);
+          else if (!strcmp (param, "file"))
+            {
+              d->args.detail = 0;
+              DETAIL_SET (d->args.detail, DETAIL_FILE_INFO);
+            }
+          else if (!strcmp (param, "elf"))
+            {
+              DETAIL_SET (d->args.detail, DETAIL_FILE_INFO);
+
+              if (!d->args.fs_root && d->srv_sckt >= 0)
+                g_print ("  You must specify root for local filesystem to "
+                         "access elf internals\n");
+              else
+                DETAIL_SET (d->args.detail, DETAIL_ELF_INFO);
+            }
           else
             {
-              g_print (emsg[EMSG_INVALID_PARAMETER], p);
+              g_print (emsg[EMSG_INVALID_PARAMETER], param);
               retval = EMSG_ERROR_NO_MESSAGE;
             }
         }
@@ -673,6 +744,13 @@ interactive_parse_line (data_t *d, char * line)
         retval = EMSG_UNKNOWN_COMMAND;
       break;
 
+    case 'f':
+      if (!strncmp (cmd, "format", cmd_len))
+        retval = interactive_format (d, param);
+      else
+        retval = EMSG_UNKNOWN_COMMAND;
+      break;
+      
     case 'h':
       if (!strncmp (cmd, "help", cmd_len))
         retval = interactive_help ();
@@ -693,26 +771,37 @@ interactive_parse_line (data_t *d, char * line)
 
     case 'p':
       if (!strcmp (cmd, "ps"))
-        retval = interactive_spawn (d, "/bin/ps", p);
+        retval = interactive_spawn (d, "/bin/ps", param);
       else if (!strncmp (cmd, "print", cmd_len))
-        retval = interactive_print (d, p);
+        retval = interactive_print (d, param);
       else
         retval = EMSG_UNKNOWN_COMMAND;
       break;
 
     case 'q':
       if (!strncmp (cmd, "quit", cmd_len))
-        return TRUE;
+        {
+          if (d->srv_sckt >= 0)
+            {
+              close (d->srv_sckt);
+            }
+          return TRUE;
+        }
       else
         retval = EMSG_UNKNOWN_COMMAND;
 
     case 'r':
       if (!strcmp (cmd, "rm"))
-        retval = interactive_rm (d, p);
+        retval = interactive_rm (d, param);
       else if (!strncmp (cmd, "reload", cmd_len))
         {
           main_reset (d);
           main_load (d);
+          retval = EMSG_SUCCESS;
+        }
+      else if (!strncmp (cmd, "root", cmd_len))
+        {
+          retval = interactive_root (d, param);
         }
       else
         retval = EMSG_UNKNOWN_COMMAND;
@@ -721,33 +810,33 @@ interactive_parse_line (data_t *d, char * line)
     case 's':
       if (!strncmp (cmd, "sort", cmd_len))
         {
-          if (!strcmp (p, "asc"))
+          if (!strcmp (param, "asc"))
             SORT_TYPE_SET (d->args.sort, SORT_TYPE_ASC);
-          else if (!strcmp (p, "desc"))
+          else if (!strcmp (param, "desc"))
             SORT_TYPE_SET (d->args.sort, SORT_TYPE_DESC);
-          else if (!strcmp (p, "vm"))
+          else if (!strcmp (param, "vm"))
             SORT_BY_SET (d->args.sort, SORT_BY_VM);
-          else if (!strcmp (p, "sole"))
+          else if (!strcmp (param, "sole"))
             SORT_BY_SET (d->args.sort, SORT_BY_SOLE);
-          else if (!strcmp (p, "heap"))
+          else if (!strcmp (param, "heap"))
             SORT_BY_SET (d->args.sort, SORT_BY_HEAP);
-          else if (!strcmp (p, "stack"))
+          else if (!strcmp (param, "stack"))
             SORT_BY_SET (d->args.sort, SORT_BY_STACK);
-          else if (!strcmp (p, "mapped"))
+          else if (!strcmp (param, "mapped"))
             SORT_BY_SET (d->args.sort, SORT_BY_MAPPED);
-          else if (!strcmp (p, "emapped"))
+          else if (!strcmp (param, "emapped"))
             SORT_BY_SET (d->args.sort, SORT_BY_EMAPPED);
 
           main_sort (d);
         }
       else if (!strncmp (cmd, "shell", cmd_len))
         {
-          gchar * s = g_strconcat ("\"", p, "\"", NULL);
+          gchar * s = g_strconcat ("\"", param, "\"", NULL);
           retval = interactive_spawn (d, "/bin/sh -r -c", s);
           g_free (s);
         }
       else if (!strncmp (cmd, "save", cmd_len))
-        retval = interactive_save (d, p);
+        retval = interactive_save (d, param);
       else
         retval = EMSG_UNKNOWN_COMMAND;
       break;
@@ -755,10 +844,7 @@ interactive_parse_line (data_t *d, char * line)
     case 't':
       if (!strncmp (cmd, "threshold", cmd_len))
         {
-          if (*p)
-            d->args.threshold = atoi (p);
-          else
-            retval = EMSG_INVALID_PARAMETER;
+          retval = interactive_threshold (d, param);
         }
     }
 
@@ -767,7 +853,7 @@ interactive_parse_line (data_t *d, char * line)
       switch (retval)
         {
         case EMSG_INVALID_PARAMETER:
-          g_print (emsg[retval], p);
+          g_print (emsg[retval], param);
           break;
 
         case EMSG_NO_PROCESS_SELECTED:
@@ -817,13 +903,24 @@ interactive_parse_line (data_t *d, char * line)
   return FALSE;
 }
 
+void
+interactive_save_history (data_t *d)
+{
+  gchar * s = get_hist_file ();
+  write_history (s);
+  g_free (s);
+}
+
 void
 interactive_loop (data_t * d)
 {
   char * line;
 
-  g_print ("  Use 'connect' command to connect to remote server or "
-           "'?' for help.\n");
+  if (d->args.remote)
+    {
+      g_print ("  Use 'connect' command to connect to remote server or "
+               "'?' for help.\n");
+    }
   
   while ((line = readline (": ")))
     {
@@ -833,7 +930,5 @@ interactive_loop (data_t * d)
       free (line);
     }
 
-  gchar * s = get_hist_file ();
-  write_history (s);
-  g_free (s);
+  interactive_save_history (d);
 }
